from __future__ import division, print_function

import myvariant
import os
import sys
import tqdm
from multiprocessing import Pool
from multiprocessing.dummy import Pool as ThreadPool
from pymongo import MongoClient
from src.base import AnnotationProject
from src.models import TxtParser, HgvsParser
import logging
logger = logging.getLogger()
logger.setLevel(logging.INFO)
logger.handlers[0].stream = sys.stdout


class VariantParsing(AnnotationProject):

    def __init__(self,
                 input_dir,
                 output_csv_path,
                 annovar_path,
                 project_data,
                 design_file=None,
                 build_ver=None):

        super(AnnotationProject, self).__init__(input_dir,
                                                output_csv_path,
                                                annovar_path,
                                                project_data,
                                                design_file=design_file,
                                                build_ver=build_ver)

        self.chunksize = 950
        self.step = 0
        self.csvs, self.vcfs = self.get_file_names()
        self.collection = project_data['project_name']
        self.db = project_data['db_name']
        self._buffer_len = 40000
        self._last_round = False

    def annotate_and_save(self, buffer=False):

        for csv, vcf in list(zip(self.csvs, self.vcfs)):

            csv_parsing = TxtParser(csv)
            hgvs = HgvsParser(vcf)
            sample_id = os.path.splitext(os.path.basename(vcf))[0]

            if not csv:

                while hgvs.num_lines > self.step * self.chunksize:

                    list_hgvs_ids = hgvs.get_variants_from_vcf(self.step)
                    myvariants_variants = self.get_dict_myvariant(list_hgvs_ids, sample_id)

                    if len(myvariants_variants) < self.chunksize:
                        self._last_round = True

                    if self._last_round:
                        return 'Done'
                    else:
                        self.export(myvariants_variants)
                        self.step += 1

                return 'Done'

            else:

                if buffer:
                    variant_buffer = []
                    while csv_parsing.num_lines > self.step * self.chunksize:

                        list_hgvs_ids = hgvs.get_variants_from_vcf(self.step)
                        myvariants_variants = self.get_dict_myvariant(list_hgvs_ids, sample_id)
                        offset = len(list_hgvs_ids) - self.chunksize
                        csv_variants = csv_parsing.open_and_parse_chunks(self.step, build_ver=self.buildver, offset=offset)

                        merged_list = []
                        for i, _ in enumerate(myvariants_variants):
                            merged_list.append(self.merge_dict_lists(myvariants_variants[i], csv_variants[i]))

                        variant_buffer.extend(merged_list)
                        self.step += 1

                        if len(merged_list) < self.chunksize:
                            self._last_round = True

                        if (len(variant_buffer) > self._buffer_len) or self._last_round:
                            logging.info('Parsing Buffer...')
                            self.export(variant_buffer)
                            variant_buffer = []

                            if self._last_round:
                                return 'Done2'

                    return 'Done1'

                else:

                    while csv_parsing.num_lines > self.step*self.chunksize:

                        list_hgvs_ids = hgvs.get_variants_from_vcf(self.step)
                        myvariants_variants = self.get_dict_myvariant(list_hgvs_ids, sample_id)
                        offset = len(list_hgvs_ids) - self.chunksize
                        csv_variants = csv_parsing.open_and_parse_chunks(self.step, offset=offset)

                        merged_list = []
                        for i, _ in enumerate(myvariants_variants):
                            merged_list.append(self.merge_dict_lists(myvariants_variants[i], csv_variants[i]))

                        if len(merged_list) < self.chunksize:
                            self._last_round = True

                        if self._last_round:
                            return 'Done'
                        else:
                            self.export(merged_list)
                            self.step += 1

                return 'Done'

    def get_file_names(self):
        vcfs = [i[0] for i in self.mapping]
        inits = [os.path.basename(i[1]) for i in self.mapping]
        csvs = []
        all_csvs = os.listdir(self.output_csv_path)
        for csv in all_csvs:
            for csv_init in inits:
                if csv.startswith(csv_init):
                    csvs.append(os.path.join(self.output_csv_path, csv))

        return vcfs, csvs

    def check_ver(self, build_ver):
        """ Checking user input validity for genome build version """

        if not build_ver:
            return 'hg19'  # Default genome build vesion

        if build_ver not in self.supported_build_vers:
            raise ValueError('Build version must not recognized. Supported builds are'
                             ' %s, %s, %s' % (self.supported_build_vers[0],
                                              self.supported_build_vers[1],
                                              self.supported_build_vers[2]))
        else:
            return build_ver

    def export(self, list_docs):
        """
        Export data do a MongoDB server
        :param list_docs: list of dictionaries containing variant information
        :return: null
        """
        client = MongoClient()
        db = getattr(client, self.db)
        collection = getattr(db, self.collection)
        collection.insert_many(list_docs, ordered=False)

    @staticmethod
    def merge_dict_lists(*dict_args):
        """
        Given any number of dicts, shallow copy and merge into a new dict,
        precedence goes to key value pairs in latter dicts.
        """
        result = {}
        for dictionary in dict_args:
            result.update(dictionary)
        return result

    def get_dict_mv_parallel(self, mapping):
        """
        Function designated to place the queries on myvariant.info servers.
        """
        sample_id = mapping[0]
        ids = mapping[1]
        mv = myvariant.MyVariantInfo()
        # This will retrieve a list of dictionaries
        variant_data = mv.getvariants(ids, as_dataframe=False)
        variant_data = self.remove_id_key(variant_data, sample_id)

        return variant_data

    def get_dict_myvariant(self, variant_list, sample_id):
        """
        Function designated to place the queries on myvariant.info servers.

        :param variant_list: list of HGVS variant ID's. Usually retrived beforehand using the method
        get_variants_from_vcf
        from the class VariantParsing.
        :return: list of dictionaries. Each dictionary contains data about a single variant.
        """

        mv = myvariant.MyVariantInfo()
        # This will retrieve a list of dictionaries
        variant_data = mv.getvariants(variant_list, as_dataframe=False)
        variant_data = self.remove_id_key(variant_data, sample_id)

        return variant_data

    def remove_id_key(self, variant_data, sample_id):

        for dic in variant_data:
            dic['hgvs_id'] = dic.pop("_id", None)
            dic['hgvs_id'] = dic.pop("query", None)
            dic['sample_id'] = sample_id

        return variant_data

    def add_csv_to_mapping(self):

        inits = self.mapping.keys()
        all_csvs = os.listdir(self.output_csv_path)
        list_of_dicts = []
        for csv in all_csvs:
            for init in inits:
                if all([csv.startswith(init), csv.endswith('txt')]):
                    self.mapping[init]['csv'] = os.path.join(self.output_csv_path, csv)
                    # list_of_dicts.append({init: self.mapping.values()})
        for key in self.mapping.keys():
            list_of_dicts.append({key: self.mapping[key]})
        return list_of_dicts

    def parallel_annotation(self, n_processes):

        list_tupls = []
        mapping = self.add_csv_to_mapping()
        for k in mapping:
            key = list(k.keys())[0]
            list_tupls.append((key, k[key]['vcf'], k[key]['csv']))

        self.pooling(n_processes, self._variant_parsing, list_tupls)

    def _variant_parsing(self, maps):

        hgvs = HgvsParser(maps[1])
        csv_parsing = TxtParser(maps[2])

        variant_buffer = []
        while csv_parsing.num_lines > self.step * self.chunksize:
            # print('LINE EXIT COND: %i, %i' % (csv_parsing.num_lines, self.step * self.chunksize))
            # logging.info('Parsing %i/%i variants from file %s' % ((self.step + 1) * self.chunksize,
            #                                                      csv_parsing.num_lines,
            #                                                      os.path.basename(maps[1])))

            list_hgvs_ids = hgvs.get_variants_from_vcf(self.step)
            # tpl = (maps[0], list_hgvs_ids)

            myvariants_variants = self.get_dict_myvariant(list_hgvs_ids, maps[0]) #self.threading(8, self.get_dict_mv_parallel, tpl)
            offset = len(list_hgvs_ids) - self.chunksize
            csv_variants = csv_parsing.open_and_parse_chunks(self.step, build_ver=self.buildver, offset=offset)

            merged_list = []
            for i, _ in enumerate(myvariants_variants):
                merged_list.append(self.merge_dict_lists(myvariants_variants[i], csv_variants[i]))

            variant_buffer.extend(merged_list)
            self.step += 1
            # print(self._last_round)

            if len(merged_list) < self.chunksize:
                self._last_round = True

            # print(self._last_round)
            if (len(variant_buffer) > self._buffer_len) or self._last_round:
                logging.info('Parsing Buffer...')
                self.export(variant_buffer)
                variant_buffer = []

                if self._last_round:
                    return 'Done2'
            # print('LINE EXIT COND: %i, %i' % (csv_parsing.num_lines, self.step * self.chunksize))

        return 'Done1'

    @staticmethod
    def threading(n_threads, input_1, input_2):

        pool = ThreadPool(n_threads)
        results = pool.map(input_1, input_2)
        pool.close()
        pool.join()
        return results

    @staticmethod
    def pooling(n_processes, input_1, input_2):

        pool = Pool(processes=n_processes)
        # rs = pool.imap(input_1, input_2)
        for _ in tqdm.tqdm(pool.imap_unordered(input_1, input_2), total=len(input_2)):
            pass



import os
import sys
import shlex
import glob
import csv
import time
import datetime
import subprocess
from collections import OrderedDict
import logging
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer
from src.base import AnnotationProject
logger = logging.getLogger()
logger.setLevel(logging.INFO)
logger.handlers[0].stream = sys.stdout


class AnnovarWrapper(AnnotationProject):

    """ Wrapper around Annovar download and annotation functions """

    def __init__(self,
                 input_dir,
                 output_csv_path,
                 annovar_path,
                 project_data,
                 design_file=None,
                 build_ver=None):

        super(AnnotationProject, self).__init__(input_dir,
                                                output_csv_path,
                                                annovar_path,
                                                project_data,
                                                design_file=design_file,
                                                build_ver=build_ver)

        self.down_dd = '-webfrom annovar'
        self.annovar_hosted = OrderedDict({'knownGene': True,
                                           'tfbsConsSites': False,
                                           'cytoBand': False,
                                           'targetScanS': False,
                                           'genomicSuperDups': False,
                                           'esp6500siv2_all': True,
                                           '1000g2015aug': True,
                                           'popfreq_all_20150413': True,
                                           'clinvar_20161128': True,
                                           'cosmic70': True,
                                           'nci60': True,
                                           'avdblist': True})

        self.dl_list_command = 'avdblist'
        self.manual_update = {'clinvar_20161128': [datetime.datetime(2016, 11, 28)],
                              '1000g2015aug':  [datetime.datetime(2016, 8, 30)],
                              'popfreq_all_20150413': [datetime.datetime(2015, 4, 13)]
                              }

        self.hg_18_databases = OrderedDict({'knownGene': 'g',
                                            'tfbsConsSites': 'r',
                                            'cytoBand': 'r',
                                            'targetScanS': 'r',
                                            'genomicSuperDups': 'r',
                                            'esp6500siv2_all': 'f',
                                            '1000g2015aug': 'f',
                                            'cosmic70': 'f',
                                            'nci60': 'f',
                                            })

        self.hg_19_databases = OrderedDict({'knownGene': 'g',
                                            'tfbsConsSites': 'r',
                                            'cytoBand': 'r',
                                            'targetScanS': 'r',
                                            'genomicSuperDups': 'r',
                                            'esp6500siv2_all': 'f',
                                            '1000g2015aug': 'f',
                                            'popfreq_all_20150413': 'f',
                                            'clinvar_20161128': 'f',
                                            'cosmic70': 'f',
                                            'nci60': 'f',
                                            })

        self.hg_38_databases = OrderedDict({'knownGene': 'g',
                                            'cytoBand': 'r',
                                            'genomicSuperDups': 'r',
                                            'esp6500siv2_all': 'f',
                                            '1000g2015aug': 'f',
                                            'clinvar_20161128': 'f',
                                            'cosmic70': 'f',
                                            'nci60': 'f',
                                            })

        self.databases = self.get_databases()

    def download_dbs(self, all_dbs=True, dbs=None):

        if len(os.listdir(os.path.join(self.annovar, 'humandb/'))) > 0:
            files = glob.glob(os.path.join(self.annovar, 'humandb/*'))
            for f in files:
                os.remove(f)

        list_commands = self.build_db_dl_command_str(all_dbs, dbs)
        for command in list_commands:
            args = shlex.split(command)
            subprocess.Popen(args, stdout=subprocess.PIPE)
            run_handler(os.path.join(self.annovar, 'humandb/'), cmds=list_commands, annovar_path=self.annovar)

        return 'Finished downloading databases to {}'.format(os.path.join(self.annovar, 'humandb/'))

    def run_annovar(self):
        """ Spawning Annovar jobs """

        n_commands = 0
        for sample in self.mapping.keys():
            annotation_dir = os.path.join(self.output_csv_path, sample)
            if os.path.isdir(annotation_dir):
                raise IsADirectoryError('Directory already exists for %s. Aborting.' % annotation_dir)
            else:
                os.makedirs(annotation_dir)

            # print(self.mapping[sample]['vcf_csv'])

            vcf_paths, csv_paths = [], []
            for vcf, csv in self.mapping[sample]['vcf_csv']:
                vcf_path = os.path.join(self.input_dir, os.path.join(sample, vcf))
                csv_path = os.path.join(self.output_csv_path, os.path.join(sample, csv))
                cmd_string = self.build_annovar_command_str(vcf_path, csv_path)
                args = shlex.split(cmd_string)
                n_commands += len(self.mapping[sample]['vcf_csv'])
                subprocess.Popen(args, stdout=subprocess.PIPE)

            logging.info('Annovar jobs submitted for sample %s' % sample)
            listen(self.output_csv_path, n_commands)
            logging.info('Finished running Annovar on sample %s' % sample)

    def build_annovar_command_str(self, vcf, csv):

        dbs = ",".join(list(self.databases.keys()))
        dbs_args = ",".join(list(self.databases.values()))

        if '1000g2015aug' in dbs:
            dbs = dbs.replace('1000g2015aug', '1000g2015aug_all')
        command = " ".join(['perl', os.path.join(self.annovar, 'table_annovar.pl'), vcf,
                            os.path.join(self.annovar, 'humandb/'), '-buildver', self.buildver, '-out',
                            csv, '-remove -protocol', dbs,  '-operation',
                            dbs_args, '-nastring .', '-otherinfo -vcfinput'])

        return command

    def build_db_dl_command_str(self, all_dbs, dbs):

        if not all_dbs:
            for db in dbs:
                if db not in self.databases:
                    raise ValueError('Database %s not supported for buil version %s' % (db, self.buildver))
            self.databases = {db: self.databases[db] for db in dbs}

        command_list = []

        for db in self.databases:

            if self.annovar_hosted[db]:
                command_list.append(" ".join(['perl', self.annovar + 'annotate_variation.pl', '-build', self.buildver,
                                              '-downdb', self.down_dd, db, self.annovar + 'humandb/']))
            else:
                command_list.append(" ".join(['perl', self.annovar + 'annotate_variation.pl', '-build', self.buildver,
                                              '-downdb', db, self.annovar + 'humandb/']))
        return command_list

    def check_for_database_updates(self):

        self.download_dbs(all_dbs=False, dbs=['avdblist'])

        with open(os.path.join(self.annovar, '/humandb' + self.buildver + '_avdblist.txt'), 'r') as db_list:
            reader = csv.reader(db_list, delimiter='\t')
            db_dict = {}
            for i in reader:

                if i[0][-6:] != 'idx.gz':
                    db_dict[i[0][5:]] = [datetime.datetime(int(i[1][0:4]), int(i[1][4:6]), int(i[1][6:8])), i[2]]

        for db in self.manual_update.keys():
            for db_ in db_dict.keys():
                if db_.startswith(db):
                    if db_dict[db_][0] > self.manual_update[db][0]:
                        logging.info('Database %s outdated, will download newer version' % db_)
                        self.download_dbs(all_dbs=False, dbs=[os.path.splitext(os.path.splitext(db_)[0])[0]])

    def get_databases(self):

        if self.buildver == 'hg18':
            databases = self.hg_18_databases
        elif self.buildver == 'hg19':
            databases = self.hg_19_databases
        else:
            databases = self.hg_38_databases

        return databases


def listen(out_path, n_files):

    added = 0

    while True:

        files = os.listdir(out_path)
        txts = sorted([i for i in files if i.endswith('txt')])

        time.sleep(5)

        new_files = os.listdir(out_path)
        new_txts = sorted([i for i in new_files if i.endswith('txt')])

        newly_created = [x for x in new_txts if x not in txts]

        if len(newly_created) > 0:
            added += 1
            logging.info('File %i/%i: Annovar finished working on file : ' % (added, n_files) +
                         os.path.basename(newly_created[0]) +
                         '.\n A text file has been created in the %s directory\n' % out_path)

        if added == n_files:
            return None


class MyHandler(FileSystemEventHandler):
    """
    Overwrite the methods for creation of files as annovar runs. Once the .csv file is detected, exit process
    and proceed to next file.
    """

    def __init__(self, observer, cmds=None, annovar_path='ANNOVAR_PATH'):

        self.observer = observer
        self.annovar = annovar_path
        self.cmds = cmds

    def on_created(self, event):

        logging.info('Downloading: ' + event.src_path)

        if self.cmds[-1][-2] in event.src_path:
            self.observer.stop()


def run_handler(output_csv_path, cmds=None, annovar_path='ANNOVAR_PATH'):
    observer = Observer()
    event_handler = MyHandler(observer,
                              cmds=cmds,
                              annovar_path=annovar_path)

    observer.schedule(event_handler, output_csv_path, recursive=True)
    observer.start()
    observer.join()

    return None
